#! /usr/bin/env python
# -*- coding: UTF-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authform.py
#
# service web de connexion au service sso Eole
# 
# compatible avec les versions 1 et 2 de la librairie CAS
# (http://www.ja-sig.org/products/cas/), a l'exception de la gestion
# des tickets de proxy (PGT)
#
###########################################################################

from socket import error as socket_err
import sys, xmlrpclib
import urlparse
from twisted.web import static
from twisted.web import resource, server, util
from config import AUTH_SERVER, AUTH_FORM_PORT, AUTH_FORM_URL, SESSION_TIMER, CERTFILE, CAS_VERSION, encoding
from eolesso.libsecure import ServerContextFactory
from eolesso import libssoclient as sso
from cgi import escape

try:
    _ = _
except NameError:
    _ = unicode

def urljoin(url, params):
    """ajoute des paramètres à une url
    >>> urljoin('http://test/','a=b')
    'http://test/?a=b'
    >>> urljoin('http://test/?a=b','c=d')
    'http://test/?a=b&c=d'
    """
    urllist = list(urlparse.urlsplit(url))
    if urllist[-2] == '':
        urllist[-2] = params
    else:
        urllist[-2] += '&' + params
    return urlparse.urlunsplit(urllist)

head = """<head>
<meta http-equiv="Pragma" content="no-cache">
<meta http-equiv="Cache-Control" content="no-store">
<meta http-equiv='Expires' content='Mon, 26 Jul 1997 05:00:00 GMT'>
<title>Authentification</title>
<link rel='stylesheet' type='text/css' href='leaves_css'>
</head>
"""

class AuthenticationVerify(resource.Resource):
    """ Page de vérification de l'identité 
        démarre une session et stocke un cookie dans le client
        si les identifiants sont corrects"""
    isLeaf = False
    addSlash = True

    def __init__(self, ip_list):
        resource.Resource.__init__(self)
        self.ip_list = ip_list

    def getChild(self, name, request):
        try:
            username = unicode(request.args['username'][0],encoding)
            password = unicode(request.args['password'][0],encoding)
            from_url = escape(request.args['service'][0])
        except:
            return
        #on va tenter l'authentification sur tous les serveurs d'auth 
        for serverip in self.ip_list:
            authserver = xmlrpclib.ServerProxy(serverip)
            session_id = authserver.authenticate(username, password)
            #authentification distante (SSO) reussie
            if session_id != "":
                # on stocke un cookie dans le navigateur
                chaine = '%s,%s'%(session_id, username.encode(encoding))
                request.addCookie("EoleSSOServer", chaine.encode(encoding), path = "/", max_age=SESSION_TIMER)
                #on redirige sur la page de l'EAD
                app_ticket = authserver.get_app_ticket(session_id, from_url)
                return util.Redirect(urljoin(from_url, 'ticket=%s' % app_ticket))
    
        # aucun des serveurs n'a authentifié l'utilisateur (meme l'authentification locale), on revient
        # sur la page de saisie de username/password
        return util.Redirect('%s?service=%s&failed=1' %(AUTH_FORM_URL, from_url))

    def render(self,request):
        return "<html>%s<body>Accès non autorisé !</body></html>"%head
    

class CASCompliantResponse(resource.Resource):
    """implémentation du protocole CAS pour la validation de ticket
    """
    def __init__(self, ip_list):
        resource.Resource.__init__(self)
        self.ip_list = ip_list

    def render(self, request):
        try:
            from_url = escape(request.args['service'][0])
            ticket = request.args['ticket'][0]
        except (IndexError, KeyError):
            return 'no\r\n\r\n'
        #on va tenter l'authentification sur tous les serveurs d'auth 
        for serverip in self.ip_list:
            authserver = xmlrpclib.ServerProxy(serverip)
            ok, infos = authserver.get_user_details(ticket, from_url)
            #authentification distante (SSO) reussie
            if ok:
                # formatting response
                if CAS_VERSION == 2:
                    response = """<cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
                        <cas:authenticationSuccess>
                        <cas:user>%s</cas:user>
                        </cas:authenticationSuccess>
                    </cas:serviceResponse>""" % infos['uid'][0]
                else:
                    response = 'yes\n%s\n' % infos['uid'][0]
                return response # dépendant structure LDAP
        # aucun des serveurs n'a authentifié l'utilisateur (meme l'authentification locale), on revient
        # sur la page de saisie de username/password
        if CAS_VERSION == 2:
            response = """<cas:serviceResponse xmlns:cas='http://www.yale.edu/tp/cas'>
                <cas:authenticationFailure code="...">
                Failed to verify session
                </cas:authenticationFailure>
            </cas:serviceResponse>"""
        else:
            response = 'no\n\n'
        return "<html>%s<body>%s</body></html>" % (head, response)

class LogoutForm(resource.Resource):
    """ 
    page de confirmation de logout (si pas d'url de redirection fournie
    """
    addSlash = True

    def render(self,request):
        return "Fin de la session SSO"

class TestForm(resource.Resource):
    """The resource that is returned when you are not logged in"""
    addSlash = True
    
    def __init__(self, *args):
        self.auth_server = xmlrpclib.ServerProxy(AUTH_SERVER)
        self.user_session = ''
        resource.Resource.__init__(self, *args)

    def is_true(self, param):
        return param == "true"
        
    def beforeRender(self, request):
        """
        protocole CAS:
         1. service : l'URL d'où on vient
         2. renew=true : on force à se réauthentifier
         3. gateway=true : si pas authentifié, on revient à l'URL
            définie par service sans ticket d'application
        """
        self.content = ""
        gateway = self.is_true(request.args.get('gateway', ['false'])[0])
        renew = self.is_true(request.args.get('renew', ['false'])[0])
        warn = self.is_true(request.args.get('warn', ['false'])[0])
        try:
            from_url = escape(request.args['service'][0])
        except KeyError:
            # nécessaire car les ressources ne sont instanciées qu'une
            # seule fois
            from_url = ''
        #on va verifier si on a un identifiant de session dans le cookie
        self.user_session = sso.get_session_id(request)
        # XXX: IE et Opera ne gerent pas les cookies de la meme
        # maniere que Firefox sous firefox, quand un cookie devient
        # invalide (au niveau du temps), il est supprimé sous IE et
        # Opera, ils sont gardés, on doit donc tester si ils ont une
        # valeur spéciale, ici foo
        
        #si on en a bien un, on le verifie vers le serveur d'authentification
        if not renew and self.user_session and self.user_session != 'foo':
            session_id, username = self.user_session.split(',')
            verif =  self.auth_server.verify_session_id(session_id.encode(encoding), username.encode(encoding))
            # si l'ID de session est verifie, on génère le ticket applicatif,
            # et on redirige vers l'URL de base
            if verif[0]:
                if from_url:
                    app_ticket = self.auth_server.get_app_ticket(session_id, from_url)
                    url = urljoin(from_url, 'ticket=%s' % app_ticket)
                    if warn:
                        self.content = """<html>%s<body onload="alert('%s'); document.location.href='%s'"></body></html>""" % (head, 'you are being authenticated', url)
                        return
                    if from_url == '':
                        
                        self.content = "<html>%s<body>pas de destination connue :)</body></html>"%head
                        return
                    return request.redirect(url)
            
                self.content = "<html>%s<body><h1>%s</h1></body></html>" % (head, 'already authenticated')
                return               
            
        if gateway and from_url:
            return request.redirect(urljoin(from_url, 'ticket='))
        # FIXME : utiliser directement la fenêtre d'authentification du navigateur ? (évite d'avoir un formulaire 'hackable')
        self.auth_msg = "Veuillez vous authentifier"
        if request.args.has_key('failed'):
            if request.args['failed'][0] == '1':
                self.auth_msg = "Echec de l'authentification, recommencez"
        self.content = """
        <html>
        %s
        <body>
        <div id="container">
           <div id="sitename">
              <h1>%s</h1>
           </div>
           <div id="wrap1">
           <div id="wrap2">
              <div id="content">
                 <form action='/verify/?service="%s"' method='post'>
                    <input type="hidden" name="service" value="%s"/>
                    <table>
                       <tr><td>Login : </td><td><input type='text' name='username'/></td></tr>
                       <tr><td>Mot de passe : </td><td><input type='password' name='password'/></td></tr>
                    </table>
                    <input type='submit'/>
                 </form>
              </div>
           </div>
           </div>
        </div>
        </body>
        </html>
        """ % (head, self.auth_msg, from_url, from_url)

    #def child_verify(self,ctx):
    #    return AuthenticationVerify([AUTH_SERVER,])

    #def child_validate(self, ctx):
    #    return CASCompliantResponse([AUTH_SERVER,])
    #
    #def child_login(self, ctx):
    #    return self
    #
    #def child_logout(self, ctx):
    #    """
    #    pour se delogger
    #    """
    #    request = inevow.IRequest(ctx)
    #    #on invalide le cookie
    #    # invlidation de la session sur le serveur d'authentification
    #    session_id = request.getCookie("EoleSSOServer")
    #    if session_id is not None:
    #        self.auth_server.logout(session_id.encode(encoding))
    #    request.addCookie("EoleSSOServer",'foo', path = "/", max_age=0)
    #    if request.args.has_key('service'):
    #        url_from = request.args['service'][0]
    #        return util.Redirect(url_from)
    #    else:
    #        #si pas de parametre from, on affiche la page de logout en local
    #        return LogoutForm()

    def getChild(self, name, request):
        if static.isDangerous(name):
            return static.DangerousPathError
        if name == 'verify':
            return AuthenticationVerify([AUTH_SERVER,])
        # FIXME : CAS 2.0
        if name == 'validate' or name == 'proxyValidate' or name == 'serviceValidate':
            return CASCompliantResponse([AUTH_SERVER,])
        if name == 'logout':
            #on invalide le cookie
            # invlidation de la session sur le serveur d'authentification
            session_id = request.getCookie("EoleSSOServer")
            if session_id is not None:
                self.auth_server.logout(session_id.encode(encoding))
            request.addCookie("EoleSSOServer",'foo', path = "/", max_age=0)
            if request.args.has_key('service'):
                url_from = escape(request.args['service'][0])
                return util.Redirect(url_from)
            else:
                #si pas de parametre from, on affiche la page de logout en local
                return LogoutForm()
        # cas "/" et /login
        self.beforeRender(request)
        return self

    def render(self,request):
        return self.content
        #return "paramètres invalides !"

def run():
    if '--test' in sys.argv:
        import doctest
        doctest.testmod(verbose=True)
    else:
        from twisted.internet import reactor
        root = TestForm()
        root.putChild('leaves_css', static.File('interface/leaves.css'))
        root.putChild('images', static.File('interface/images'))
        reactor.listenSSL(int(AUTH_FORM_PORT), server.Site(root), ServerContextFactory(CERTFILE))
        reactor.run()

if __name__ == '__main__':
    run()
