#! /usr/bin/env python
# -*- coding: UTF-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authserver.py
#
# serveur xmlrpc de gestion de session SSO Eole
#
###########################################################################

import time, socket
from xmlrpclib import ServerProxy
import ldap
# imports twisted
from twisted.internet import reactor
from twisted.web import resource, server, util
from twisted.web.xmlrpc import XMLRPC
# imports sso eole
from config import SESSION_TIMER, APP_TIMER, LDAP_SERVER, LDAP_BASE,AUTH_SERVER_PORT, CERTFILE, encoding
from eolesso import ticketcache
from eolesso.libsecure import ServerContextFactory

class IAuthServer:
    def authenticate(username, password):
        """ """

    def serverid():
        """ """

def verif_cert(conn, cert, err_num, err_depth, rc):
    """fonction de vérification du certificat"""
    return not cert.has_expired()

class AuthServer(XMLRPC):
    __implements__ = (IAuthServer)

    def __init__(self,data_proxy,parent=None):
        XMLRPC.__init__(self)
        self.address = socket.getfqdn()
        # cache des "login tickets"
        self.user_sessions = ticketcache.TicketCache(SESSION_TIMER, self.address, renew_tickets=True)
        # cache des "application tickets"
        self.app_sessions = ticketcache.TicketCache(APP_TIMER, self.address, 'ST')
        self._data_proxy = data_proxy
        self._parent = None
        if parent:
            self._parent = ServerProxy(parent)
    
    def xmlrpc_serverid(self):
        return id(self)
        
    def xmlrpc_authenticate(self, username, password):
        sessionID = ''
        if type(username) == unicode: username = username.encode(encoding)
        if type(password) == unicode: password = password.encode(encoding)

        # on demande en premier au serveur père qu'il vérifie les informations
        if self._parent:
            sessionID = self._parent.authenticate(username, password)
        if sessionID == '' and password != '':
            # non authentifié par un parent, on essaye de le faire localement
            try:
                if self._data_proxy.authenticate(username, password):
                    return self.user_sessions.add_session(username)
            except ldap.SERVER_DOWN:
                return ''
            else:
                return ''
        else:
            return sessionID

    def xmlrpc_get_app_ticket(self, session, appurl):
        # si cette session utilisateur est valide, on crée un ticket pour l'application
        AppSessionID = ''
        if self.user_sessions.validate_session(session):
            AppSessionID = self.app_sessions.add_session([appurl,session])
            return AppSessionID
        else:
            # sinon on demande  au serveur père de le créer
            if self._parent:
                AppSessionID = self._parent.get_app_ticket(session, appurl)
                return AppSessionID
            else:
                print "session not validated"
                return ''
    
    # FIXME : Récupérer les infos dans un serveur ldap ?
    def _get_user_details(self, session_id):
        """Renvoie les informations connues de l'utilisateur lié à <session_id>"""
        try:
            # on récupère l'id de l'utilisateur
            user_id = self.user_sessions.get_session_info(session_id)
            try:
                infos = self._data_proxy.get_data(user_id)
            except ldap.SERVER_DOWN:
                return {}
            if infos == {}:
                # pas d'infos locales, on regarde  sur le serveur parent
                if self._parent:
                    return self._parent.get_user_details(session_id)
                return {}
            else:
                return infos
        except ticketcache.InvalidSession:
            return {}

    def xmlrpc_verify_session_id(self, session_id, username):
        """vérifie un session_id"""
        if self.user_sessions.validate_session(session_id):
            try:
                if self.user_sessions.get_session_info(session_id) == username:
                    return True,session_id
            except:
                pass
        return False,session_id

    def xmlrpc_logout(self, session_id):
        """vérifie un session_id"""
        self.user_sessions.end_session(session_id)
        return 'ok'

    def xmlrpc_verify_app_ticket(self, app_ticket, appurl):
        """vérifie le ticket fourni par une application"""
        if self.app_sessions.validate_session(app_ticket):
            try:
                if self.app_sessions.get_session_info(app_ticket)[0] == appurl:
                    return True, app_ticket
            except:
                pass
        return False, app_ticket

    def xmlrpc_get_user_details(self, app_ticket, appurl):
        """renvoie le le login de l'utilisateur"""
        if self.app_sessions.validate_session(app_ticket):
            try:
                stored_url, session_id = self.app_sessions.get_session_info(app_ticket)
                if stored_url == appurl:
                    return True, self._get_user_details(session_id)
            except KeyError:
                print "ERR : session not found"
                # la session n'existe pas
                pass
        print "not valid session !!"
        return False, {}

        
def run(*args):
    # intialisation du serveur
    from eolesso.dataproxy import LDAPProxy
    proxy = LDAPProxy(LDAP_SERVER)
    # lancement du serveur
    from twisted.web import server
    reactor.listenSSL(int(AUTH_SERVER_PORT), server.Site(AuthServer(proxy)),ServerContextFactory(CERTFILE))
    print "Serveur initialisé (port %s)" % AUTH_SERVER_PORT
    reactor.run()


if __name__ == '__main__':
    run()
