import sys
import Cookie

from twisted.web2 import server, http, resource, channel, responsecode
from twisted.internet import reactor

from openid.server import server as oidserver
# from openid.store.memstore import MemoryStore
from openid.store.filestore import FileOpenIDStore

## mock database #####################################################
DATABASE = {
    'adim' : dict(groups=('tech', 'prestat'), nick='adim',
                  firstname='Adrien', lastname='Di Mascio'),

    'luc' : dict(groups=('tech', 'admin'), nick='luc',
                 firstname='Luc', lastname='Bourdot'),
    }




## web resources #####################################################



def makePage(body, title="zou", additionalHeaders=""):
    return """<html>
 <head>
  <title>%s</title>
  %s
 </head>
 <body>%s
 </body>
</html>
""" % (title, additionalHeaders, body)

    
class UserOwnPage(resource.Resource):
    template = """<html>
 <head>
  <title>%(nick)s's home page</title>
  <link rel="openid.server" href="%(server)s" />
 </head>
 <body>
  <h1>zou</h1>
 </body>
</html>
"""

    def __init__(self, nickname):
        resource.Resource.__init__(self, nickname)
        self.nickname = nickname

    def render(self, ctx):
        url = "%s://%s" % (ctx.scheme, ctx.host)
        if ctx.port:
            url += ':%s' % ctx.port
        # url += '/identify'
        url += '/oidendpoint'
        return http.Response(stream=self.template %
                             {'nick' : self.nickname, 'server' : url})


"""
?openid.identity=http%3A%2F%2Flocalhost%3A8080%2Fhome%2Fadim&openid.mode=checkid_setup&openid.return_to=http%3A%2F%2Flocalhost%3A8001%2Fprocess%3Fjanrain_nonce%3D2007-06-11T10%253A00%253A53ZuYIwqU&openid.trust_root=http%3A%2F%2Flocalhost%3A8001%2F
"""

class HomeDir(resource.Resource):
    addSlash = True
    
    def render(self, ctx):
	return http.Response(stream="yo yo")

    def locateChild(self, request, segments):
        """Indicate which resource to use to process down the URL's path"""
        if segments and segments[0]:
            return UserOwnPage(segments[0]), ()
        return self, ()


class AuthenticationResource(resource.Resource):
    """resource responsible for user authentication"""

    source = """
        <h2>Login</h2>
        %s
        <form method="GET" action="/authenticate">
          <input type="hidden" name="success_to" value="%s" />
          <input type="hidden" name="fail_to" value="%s" />
          <input type="text" name="user" value="" />
          <input type="password" name="password" value="" />
          <input type="submit" name="submit" value="Log In" />
          <input type="submit" name="cancel" value="Cancel" />
        </form>"""

    def render(self, ctx):
        args = ctx.args
        identifier = args['openid.identity'][0]
        checkid_mode = args['openid.mode'][0]
        return_path = args['openid.return_to'][0]
        trust_root = args['openid.trust_root'][0]
        positive_headers = {
            'openid.ns' : "http://specs.openid.net/auth/2.0",
            'openid.mode' : "id_res",
            'openid_endpoint' : "",
            
            }
        errmsg = args.get('errmsg', [''])[0]
        onfailure = '/identify?errmsg=login%%20ou%%20password%%20incorrect&openid.identify=%s&openid.mode=%s&openid.return_to=%s&openid.trust_root=%s' % (
            identifier, checkid_mode, return_path, trust_root)
        source = self.source % (errmsg, return_path, onfailure)
        return http.Response(stream=makePage(source, title="please log in"))

    

class AuthenticateUser(resource.Resource):
    
    def render(self, ctx):
        args = ctx.args
        if self._authenticate(args['user'][0],
                              args['password'][0]):
            redirect_to = args['success_to'][0] + '&openid.mode=id_res'
        else:
            redirect_to = args['fail_to'][0]
        print "redirecting to", redirect_to
        return http.RedirectResponse(redirect_to)
        

    def _authenticate(self, user, password):
        # Do something appropriate here
        return user == password


def simplifyQuery(args):
    """transforms a query dict as stored in twisted's ctx objects
    in a simple arg / value dict

    (twisted's ctx.args objects map arg to list of values)
    """
    query = {}
    for key, value in args.items():
        if type(value) is list:
            if len(value) == 1:
                query[key] = value[0]
            else:
                raise "ooppppssssssssssssssss"
                # query[key] = value
        else:
            query[key] = value
    return query


def getCookiedUser(cookies):
    print "cookies =", cookies
    if cookies:
        morsel = Cookie.BaseCookie(cookies).get('user')
        if morsel: 
            print "got morsel, value =", morsel.value
            return morsel.value
    return None
    
    
class OidServerEndPoint(resource.Resource):
    """OpenID EndPoint"""
    _allowed_methods = ('GET', 'POST')

    def render(self, ctx):
        # XXX oidserver could probably be "remembered" on the ctx
        print "*" * 40
        print ctx.args
        print "*" * 40
        self.user = None
        try:
            try:
                request = self.oidserver.decodeRequest(simplifyQuery(ctx.args))
            except oidserver.ProtocolError, why:
                print "ouch ouch ouch", why
                return http.Response(stream="ouch ouch ouch +%s" % why)
            if request is None:
                return http.Response(stream="tell something about endpoint")

            if request.mode in ("checkid_immediate", "checkid_setup"):
                return self.handleCheckIDRequest(request)
            else:
                response = self.oidserver.handleRequest(request)
                return self.displayResponse(response)
        finally:
            self.user = None
        
    
    def displayResponse(self, response):
        try:
            webresponse = self.oidserver.encodeResponse(response)
        except oidserver.EncodingError, why:
            text = why.response.encodeToKVForm()
            return http.StatusResponse(500, '<pre>%s</pre>' % text,
                                       title="encoding error")

        return http.Response(stream=webresponse.body,
                             headers=webresponse.headers)


    def handleCheckIDRequest(self, request):
        is_authorized = self.isAuthorized(request.identity, request.trust_root)
        if is_authorized:
            response = self.approved(request)
            return self.displayResponse(response)
        elif request.immediate:
            response = request.answer(False)
            return self.displayResponse(response)
        else:
            self.lastCheckIDRequest[self.user] = request
            return self.showDecidePage(request)


    def isAuthorized(self, identity_url, trust_root):
        if self.user is None:
            return False

        if identity_url != self.base_url + 'home/' + self.user:
            return False

        key = (identity_url, trust_root)
        return self.oidserver.approved.get(key) is not None


    def showDecidePage(self, request):
        id_url_base = self.base_url+'home/'
        expected_user = request.identity[len(id_url_base):]

        if request.idSelect(): # We are being asked to select an ID
            msg = '''\
            <p>A site has asked for your identity.  You may select an
            identifier by which you would like this site to know you.
            On a production site this would likely be a drop down list
            of pre-created accounts or have the facility to generate
            a random anonymous identifier.
            </p>
            '''
            fdata = {
                'id_url_base': id_url_base,
                'trust_root': request.trust_root,
                }
            form = '''\
            <form method="POST" action="/allow">
            <table>
              <tr><td>Identity:</td>
                 <td>%(id_url_base)s<input type='text' name='identifier'></td></tr>
              <tr><td>Trust Root:</td><td>%(trust_root)s</td></tr>
            </table>
            <p>Allow this authentication to proceed?</p>
            <input type="checkbox" id="remember" name="remember" value="yes"
                /><label for="remember">Remember this
                decision</label><br />
            <input type="submit" name="yes" value="yes" />
            <input type="submit" name="no" value="no" />
            </form>
            '''%fdata
        elif expected_user == self.user:
            msg = '''\
            <p>A new site has asked to confirm your identity.  If you
            approve, the site represented by the trust root below will
            be told that you control identity URL listed below. (If
            you are using a delegated identity, the site will take
            care of reversing the delegation on its own.)</p>'''

            fdata = {
                'identity': request.identity,
                'trust_root': request.trust_root,
                }
            form = '''\
            <table>
              <tr><td>Identity:</td><td>%(identity)s</td></tr>
              <tr><td>Trust Root:</td><td>%(trust_root)s</td></tr>
            </table>
            <p>Allow this authentication to proceed?</p>
            <form method="POST" action="/allow">
              <input type="checkbox" id="remember" name="remember" value="yes"
                  /><label for="remember">Remember this
                  decision</label><br />
              <input type="submit" name="yes" value="yes" />
              <input type="submit" name="no" value="no" />
            </form>''' % fdata
        else:
            mdata = {
                'expected_user': expected_user,
                'user': self.user,
                }
            msg = '''\
            <p>A site has asked for an identity belonging to
            %(expected_user)s, but you are logged in as %(user)s.  To
            log in as %(expected_user)s and approve the login request,
            hit OK below.  The "Remember this decision" checkbox
            applies only to the trust root decision.</p>''' % mdata

            fdata = {
                'identity': request.identity,
                'trust_root': request.trust_root,
                'expected_user': expected_user,
                }
            form = '''\
            <table>
              <tr><td>Identity:</td><td>%(identity)s</td></tr>
              <tr><td>Trust Root:</td><td>%(trust_root)s</td></tr>
            </table>
            <p>Allow this authentication to proceed?</p>
            <form method="POST" action="/allow">
              <input type="checkbox" id="remember" name="remember" value="yes"
                  /><label for="remember">Remember this
                  decision</label><br />
              <input type="hidden" name="login_as" value="%(expected_user)s"/>
              <input type="submit" name="yes" value="yes" />
              <input type="submit" name="no" value="no" />
            </form>''' % fdata

        return http.Response(stream=makePage(msg+form))

    
class AllowUser(resource.Resource):
    _allowed_methods = ('GET', 'POST')

    def render(self, ctx):
        self.uset = getCookiedUser(ctx.headers.getRawHeaders('Cookie'))
        query = simplifyQuery(ctx.args)
        # pretend this next bit is keying off the user's session or something,
        # right?
        request = self.lastCheckIDRequest.get(self.user)
        if 'yes' in query:
            if 'login_as' in query:
                self.user = self.query['login_as']

            trust_root = request.trust_root
            if self.query.get('remember', 'no') == 'yes':
                self.oidserver.approved[(identity, trust_root)] = 'always'

            if request.idSelect():
                identity = self.base_url + 'home/' + query['identifier']
            else:
                identity = request.identity

            response = self.approved(request, identity)

        elif 'no' in query:
            response = request.answer(False)

        else:
            assert False, 'strange allow post.  %r' % (query,)

        self.displayResponse(response)
    
    def http_POST(self, request):
        return self.render(request)


    
class Toplevel(resource.Resource):
    addSlash = True


    child_home = HomeDir()
    child_identify = AuthenticationResource()
    child_oidendpoint = OidServerEndPoint()
    child_authenticate = AuthenticateUser()
    child_allow = AllowUser()
    
    source = """<html>
  <head><title>zou</title></head>
  <body>
  <h1>Root page, go to <a href="home/adim">adim's page</a></h1>
  </body>
</html>
"""
    def __init__(self, base_url="http://localhost:8080/"):
        resource.Resource.__init__(self)
        self.base_url = base_url
        endpoint = base_url + 'oidendpoint'
        store = store = FileOpenIDStore('oid_store.dat')
        self.oidserver = oidserver.Server(store, endpoint)
        self.approved = {}
        self.lastCheckIDRequest = {}
        # XXX ouch ouch ouch (find a twistedish way to do that)
        #     maybe ctx.remember() is the correct way to do it ?
        self.child_oidendpoint.oidserver = self.oidserver
        self.child_oidendpoint.approved = self.approved
        self.child_oidendpoint.lastCheckIDRequest = self.lastCheckIDRequest
        self.child_oidendpoint.base_url = self.base_url
        
        self.child_allow.oidserver = self.oidserver
        self.child_allow.lastCheckIDRequest = self.lastCheckIDRequest
        self.child_allow.base_url = self.base_url
        
        
    def render(self, ctx):
	return http.Response(stream=self.source)
    



## main ##############################################################
def run(port=8080):
    site = server.Site(Toplevel())
    reactor.listenTCP(port, channel.HTTPFactory(site))
    print "c'est parti (port=%s)" % port
    reactor.run()

if __name__ == '__main__':
    if len(sys.argv) > 1:
        port = int(sys.argv[1])
    else:
        port = 8080
    run()
